local box2dDebug = require("libs.debugWorldDraw")
local cpml = require("libs.cpml")
local Camera = require("libs.hump.camera")
local cam
local objects = {}
local world

function love.load()
	
	love.physics.setMeter(64)
	world = love.physics.newWorld(0,0,true)
	world:setCallbacks(beginContact, endContact)

	cam = Camera(30,30)

	local groundArea = {}
	groundArea.x = 40
	groundArea.y = 40
	groundArea.width = 400
	groundArea.height = 500
	groundArea.angle = math.rad(50)
	groundArea.frictionModifier = .80
	makeGroundRect(groundArea)

	carBody = {}
	carBody.body = love.physics.newBody(world,300,300,"dynamic" )
	carBody.shape = love.physics.newPolygonShape( 22.5,0, 45,37.5, 42,82.5, 15,150, -15,150, -42,82.5, -45,37.5, -22.5,0 )
	carBody.fixture = love.physics.newFixture(carBody.body, carBody.shape)

	carBody.flTire = makeTire( carBody.body:getWorldPoint(40,130) )
	joint = love.physics.newRevoluteJoint( carBody.body, carBody.flTire.body, carBody.flTire.body:getPosition()  )
	joint:setLimits(0,0)
	joint:setLimitsEnabled(true)
	carBody.flTire.joint = joint

	carBody.frTire = makeTire( carBody.body:getWorldPoint(-40,130) )
	joint = love.physics.newRevoluteJoint( carBody.body, carBody.frTire.body, carBody.frTire.body:getPosition()  )
	joint:setLimits(0,0)
	joint:setLimitsEnabled(true)
	carBody.frTire.joint = joint

	carBody.rlTire = makeTire( carBody.body:getWorldPoint(43,9) )
	joint = love.physics.newRevoluteJoint( carBody.body, carBody.rlTire.body, carBody.rlTire.body:getPosition()  )
	joint:setLimits(0,0)
	joint:setLimitsEnabled(true)
	carBody.rlTire.joint = joint

	carBody.rrTire = makeTire( carBody.body:getWorldPoint(-43,9) )
	joint = love.physics.newRevoluteJoint( carBody.body, carBody.rrTire.body, carBody.rrTire.body:getPosition()  )
	joint:setLimits(0,0)
	joint:setLimitsEnabled(true)
	carBody.rrTire.joint = joint

end

function love.update( dt )
	world:update(dt)	

	updateDrive(carBody.flTire)
	updateTurn(carBody.flTire, dt)
	updateFriction(carBody.flTire, dt)

	updateDrive(carBody.frTire)
	updateTurn(carBody.frTire, dt)
	updateFriction(carBody.frTire, dt)

	updateDrive(carBody.rlTire)
	updateFriction(carBody.rlTire, dt)

	updateDrive(carBody.rrTire)
	updateFriction(carBody.rrTire, dt)

	cam:lookAt(carBody.body:getPosition())
end

function love.draw(  )
	cam:attach()
	box2dDebug(world, -2000,-2000, 4000,4000 )
	cam:detach()

end

function love.mousepressed( x, y, button, istouch )

end

function love.mousereleased( x, y, button, istouch )

end

function love.keypressed( key )

end

function love.keyreleased( key )

end

function makeTire( x,y )
	local tire = {}
	tire.type = "tire"
	tire.frictionModifier = 1
	tire.frictionModifierGoal = 1
	tire.updateValue = 1
	tire.body = love.physics.newBody(world,x,y,"dynamic" )
	tire.shape = love.physics.newRectangleShape(.5*25,1.25*25)
	tire.fixture = love.physics.newFixture(tire.body, tire.shape)
	tire.fixture:setUserData(tire)

	return tire
end

function makeGroundRect( properties )
	local groundObject = {}
	groundObject.x = properties.x
	groundObject.y = properties.y
	groundObject.width = properties.width
	groundObject.height = properties.height
	groundObject.angle = properties.angle
	groundObject.frictionModifier = properties.frictionModifier
	groundObject.type = "ground"

	groundObject.body = love.physics.newBody(world,groundObject.x,groundObject.y,"static")
	groundObject.shape = love.physics.newRectangleShape(groundObject.width,groundObject.height)
	groundObject.fixture = love.physics.newFixture(groundObject.body,groundObject.shape)
	groundObject.body:setAngle(groundObject.angle)

	groundObject.fixture:setUserData(groundObject)
	groundObject.fixture:setSensor(true)

	return groundObject

end

function getLateralVelocity( object )
	local currentRightNormal = cpml.vec2(object.body:getWorldVector(1,0))
	local currentLinearVelocity = cpml.vec2(object.body:getLinearVelocity())
	local dotProduct = currentRightNormal:dot(currentLinearVelocity)
	return currentRightNormal:scale(dotProduct)
end

function getForwardVelocity( object )
	local currentUpNormal = cpml.vec2(object.body:getWorldVector(0,1))
	local currentLinearVelocity = cpml.vec2(object.body:getLinearVelocity())
	local dotProduct = currentUpNormal:dot(currentLinearVelocity)
	return currentUpNormal:scale(dotProduct)
end


function updateFriction( object, dt )

	if(object.frictionModifierGoal > object.frictionModifier) then
		object.frictionModifier = object.frictionModifier + object.updateValue*dt
		if (object.frictionModifier > object.frictionModifierGoal) then
			object.frictionModifier = object.frictionModifierGoal
		end
	elseif (object.frictionModifierGoal < object.frictionModifier) then
		object.frictionModifier = object.frictionModifier + object.updateValue*dt
		if (object.frictionModifier < object.frictionModifierGoal) then
			object.frictionModifier = object.frictionModifierGoal
		end
	end

	local impulse = getLateralVelocity(object)
	local body = object.body
	impulse = impulse:scale(-body:getMass())

	if(impulse:len() > 0.01) then
		local skidModifier = object.frictionModifier^11
		impulse = impulse:scale(skidModifier/impulse:len())
	end
	body:applyLinearImpulse(impulse.x*object.frictionModifier,impulse.y*object.frictionModifier)
	body:applyAngularImpulse(0.1*body:getInertia() * -body:getAngularVelocity())

	local currentForwardSpeed = getForwardVelocity(object):normalize()
	local dragForceMagnitude = currentForwardSpeed:scale(-2)
	body:applyForce(dragForceMagnitude.x*object.frictionModifier, dragForceMagnitude.y*object.frictionModifier)


end

function updateDrive( object )
	local maxForwardSpeed   = 900
	local maxBackwardSpeed  = -900
	local maxDriveForce     = 400
	local body = object.body

	local desiredSpeed = 0
	local isDown = love.keyboard.isDown
	if isDown("w") then
		desiredSpeed = maxForwardSpeed
	elseif isDown("s") then
		desiredSpeed = maxBackwardSpeed
	end

	local currentForwardNormal = cpml.vec2(body:getWorldVector(0,1))
	local currentSpeed = currentForwardNormal:dot(getForwardVelocity(object))

	local force = 0
	if (desiredSpeed > currentSpeed) then
		force = maxDriveForce
	elseif (desiredSpeed < currentSpeed) then
		force = -maxDriveForce/2
	end

	local forceVector = currentForwardNormal:scale(force)
	body:applyForce(forceVector.x*object.frictionModifier,forceVector.y*object.frictionModifier)
end

function updateTurn( object, dt )
	local lockAngle = math.rad(40)
	local isDown = love.keyboard.isDown
	local x,y = object.joint:getLimits()
	if isDown("d") then
		x = x + (lockAngle)*dt
		if x > lockAngle then
			x = lockAngle
		end
		object.joint:setLimits(x,x)
	elseif isDown("a") then
		x = x - (lockAngle)*dt
		if x < -lockAngle then
			x = -lockAngle
		end
		object.joint:setLimits(x,x)
	else
		if x > 0 then
			x = x - lockAngle*dt
		elseif x < 0 then 
			x = x + lockAngle*dt
		end
		object.joint:setLimits(x,x)
	end
end

function collisionHandler( tire, groundArea, begin )
	local tireObj = tire:getUserData()
	local groundObj = groundArea:getUserData()
	if begin then
		tireObj.frictionModifierGoal = groundObj.frictionModifier
		tireObj.updateValue = (tireObj.frictionModifierGoal - tireObj.frictionModifier)/.25

	else
		tireObj.frictionModifierGoal = 1
		tireObj.updateValue = (tireObj.frictionModifierGoal - tireObj.frictionModifier)/.25
	end

end

function beginContact( fixtureA, fixtureB, contacts )
	if isType(fixtureA, "tire") and isType(fixtureB, "ground") then
		collisionHandler(fixtureA,fixtureB,true)
	elseif isType(fixtureA, "ground") and isType(fixtureB, "tire") then
		collisionHandler(fixtureB,fixtureA,true)
	end
end

function endContact( fixtureA, fixtureB, contacts )
	if isType(fixtureA, "tire") and isType(fixtureB, "ground") then
		collisionHandler(fixtureA,fixtureB,false)
	elseif isType(fixtureA, "ground") and isType(fixtureB, "tire") then
		collisionHandler(fixtureB,fixtureA,false)
	end
end

function isType(fixture, type)
	return (fixture:getUserData() and fixture:getUserData().type == type)
end

